#include <fggl/gfx/renderer.hpp>
#include <fggl/gfx/rendering.hpp>
#include <fggl/gfx/camera.hpp>

#include <fggl/data/model.hpp>

#include <glm/ext/matrix_transform.hpp>
#include <glm/glm.hpp>
#include <glm/gtc/type_ptr.hpp>


/**
 * Future optimisations:
 *   recommended approach is to group stuff in to as few vao as possible - this will do one vao per mesh, aka bad.
 *   Add support for instanced rendering (particles)
 *   Look at packing vertex data in better ways (with profiling)
 *   Support shader specialisation (ie, dynamic/streamed data)
 *   Follow best recommendations for Vertex attributes (ie normals not using floats, bytes for colour)
 *
 * Future features:
 *   Add support for models with weights (for animations)
 *   OpenGL ES for the FOSS thinkpad users who can't run anything even remotely modern
 */
 

using namespace fggl::gfx;

template<typename T>
static GLuint createArrayBuffer(std::vector<T>& vertexData) {
	GLuint buffId;
	glGenBuffers(1, &buffId);
	glBindBuffer(GL_ARRAY_BUFFER, buffId);
	glBufferData(GL_ARRAY_BUFFER,
			vertexData.size() * sizeof(T),
			vertexData.data(),
			GL_STATIC_DRAW);
	return buffId;	
}

static GLuint createIndexBuffer(std::vector<uint32_t>& indexData) {
	GLuint buffId;
	glGenBuffers(1, &buffId);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, buffId);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER,
			indexData.size() * sizeof(uint32_t),
			indexData.data(),
			GL_STATIC_DRAW);
	return buffId;
}

MeshRenderer::token_t MeshRenderer::upload(fggl::data::Mesh& mesh) {
	MeshRenderer::token_t token;
	glGenVertexArrays(1, &token.vao);
	glBindVertexArray(token.vao);

	token.buffs[0] = createArrayBuffer<fggl::data::Vertex>( mesh.vertexList() );
	glEnableVertexAttribArray(0);
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, sizeof(fggl::data::Vertex), reinterpret_cast<void*>(offsetof(fggl::data::Vertex, posititon)));
	glEnableVertexAttribArray(1);
	glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, sizeof(fggl::data::Vertex), reinterpret_cast<void*>(offsetof(fggl::data::Vertex, normal)));

	token.buffs[1] = createIndexBuffer( mesh.indexList() );
	token.idxOffset = 0;
	token.idxSize = mesh.indexCount();
	glBindVertexArray(0);

	return token;
}

void MeshRenderer::render(const Window& window, const fggl::ecs::ECS& ecs, const ecs::entity_t camera, float dt) {
    if ( camera == ecs::NULL_ENTITY ) return;

    auto entities = ecs.getEntityWith<MeshRenderer::token_t>();
    if ( entities.size() == 0 ) return;

    total += dt;

    // make sure the correct rendering context is active
    window.activate();

    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);

    glEnable(GL_DEPTH_TEST);

    // camera logic
    const auto camTransform = ecs.getComponent<math::Transform>(camera);
    const auto camComp = ecs.getComponent<gfx::Camera>(camera);
    glm::mat4 proj = glm::perspective( camComp->fov, camComp->aspectRatio, camComp->nearPlane, camComp->farPlane);
    glm::mat4 view = glm::lookAt( camTransform->origin(), camComp->target, camTransform->up() );

    // lighting
    glm::vec3 lightPos(20.0f, 20.0f, 15.0f);

    // TODO better performance if grouped by vao first
    // TODO the nvidia performance presentation said I shouldn't use uniforms for large data
    for ( auto& entity : entities ) {
	    const auto& transform = ecs.getComponent<fggl::math::Transform>(entity);
	    const auto& mesh = ecs.getComponent<MeshRenderer::token_t>(entity);

	    glm::mat4 model = transform->model();
//	    model = glm::rotate(model, glm::radians(total/2048.0f * 360.0f), glm::vec3(0.0f,1.0f,0.0f));

	    auto shader = mesh->pipeline;
	    glUseProgram( shader );

	    glUniformMatrix4fv( glGetUniformLocation(shader, "model"), 1, GL_FALSE, glm::value_ptr( model ) );
	    glUniformMatrix4fv( glGetUniformLocation(shader, "view"), 1, GL_FALSE, glm::value_ptr( view ) );
	    glUniformMatrix4fv( glGetUniformLocation(shader, "projection"), 1, GL_FALSE, glm::value_ptr( proj ) );

	    // lighting
	    GLint lightID = glGetUniformLocation(shader, "lightPos");
	    if ( lightID != -1 )
		    glUniform3fv( lightID, 1, glm::value_ptr( lightPos ) );

	    glBindVertexArray( mesh->vao );
	    glDrawElements( GL_TRIANGLES, mesh->idxSize, GL_UNSIGNED_INT, reinterpret_cast<void*>(mesh->idxOffset) );
    }

    glBindVertexArray(0);

}

